#
#  Copyright (c) 1997-2002 The Protein Laboratory, University of Copenhagen
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
#  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.
#
#  $Id: extlist.pl,v 1.7 2002/05/14 13:22:26 dk Exp $
#

=pod 
=item NAME

Prima check-list widget

=item FEATURES

Demonstrates basic usage of a Prima toolkit and
its standard CheckList class.


=cut

use Prima;
use Prima::ExtLists;
use Prima::Application;

my $w = Prima::Window->create(
   size => [ 200, 200],
   onDestroy => sub { $_[0]-> owner-> close; },
);
my $v = '';
vec($v, 0, 8) = 0x77;
$w-> insert( Prima::CheckList =>
   origin => [0,0],
   size   => [ $w-> size],
   growMode => gm::Client,
   items    => [qw(Tra lala tri rubla hop hey lala ley comin down the judgement day)],
   multiColumn => 1,
   multiSelect => 1,
   vector   => $v,
   extendedSelect => 0,
);

run Prima;
